"""
Copyright (C) 2024, Jabez Winston C

Embedded Tester Protocol Library - ADC

Author  : Jabez Winston C <jabezwinston@gmail.com>
License : MIT
Date    : 19-Sep-2024

"""

import struct

class ADC:
    code = 2
    ops = {
        'info': 0,
        'init': 1,
        'ctrl': 2,
        'read': 3
    }

    def __init__(self, etp):
        self.etp = etp
        self.reference_mv = 3300
        self.resolution = 10

    """
    Query ADC information
    """

    def get_info(self):
        adc_info_cmd = self.code << 8 | self.ops['info']
        cmd = self.etp.frame_packet(adc_info_cmd)
        self.etp.cmd_queue.put(cmd)
        rsp = self.etp.read_rsp()
        num_adc, num_channels, resolution, reference, max_rate, port_count = struct.unpack('<BBBHIB', rsp[0:10])
        adc_ports = list(struct.iter_unpack('<BI', rsp[10:]))
        adc_info = []
        for port in adc_ports:
            adc_info.append({'port': chr(port[0]), 'pins': self.etp.mask_to_bits(port[1], 32)})

        self.reference_mv = reference
        self.resolution = resolution

        return {
            'num_adc': num_adc,
            'num_channels': num_channels,
            'resolution': resolution,
            'reference': reference,
            'max_rate': max_rate,
            'port_count': port_count,
            'ports': adc_info
        }

    """
    Enable/Disable ADC pins

    """

    def init(self, pin_list):
        adc_init_cmd = self.code << 8 | self.ops['init']
        adc_pin_mask = 0
        adc_enable_mask = 0
        for pin in pin_list.keys():
            port, pin_num = self.etp.gpio.decode_gpio_pin(pin)
            if (pin_list[pin] == True):
                adc_enable_mask |= (1 << pin_num)
            adc_pin_mask |= (1 << pin_num)

        cmd = self.etp.frame_packet(adc_init_cmd, struct.pack('<BII', ord(port), adc_pin_mask, adc_enable_mask))
        self.etp.cmd_queue.put(cmd)

    """
    Control the ADC pin
    
    """

    def ctrl(self, pin, **kwargs):
        adc_ctrl_cmd = self.code << 8 | self.ops['ctrl']
        port, pin = self.etp.gpio.decode_gpio_pin(pin)

        if 'resolution' in kwargs:
            resolution = kwargs['resolution']
        else:
            resolution = self.resolution

        if 'reference_mv' in kwargs:
            reference_mv = kwargs['reference_mv']
        else:
            reference_mv = self.reference_mv

        if 'rate' in kwargs:
            rate = kwargs['rate']
        else:
            rate = 1000

        if rate < 65536:
            unit = 0x01
        else:
            unit = 0x02
            rate = rate // 1000

        if 'start' in kwargs:
            start = kwargs['start']
        else:
            start = 0

        cmd = self.etp.frame_packet(adc_ctrl_cmd, struct.pack('<BBBHHBB', ord(port), pin, resolution, reference_mv, rate, unit, start))
        self.etp.cmd_queue.put(cmd)
        rsp = self.etp.read_rsp()
        return rsp

    """
    Read data from ADC pin

    """

    def read(self, pin_str):
        adc_read_cmd = self.code << 8 | self.ops['read']
        port, pin = self.etp.gpio.decode_gpio_pin(pin_str)
        cmd = self.etp.frame_packet(adc_read_cmd, struct.pack('<BB', ord(port), pin))
        self.etp.cmd_queue.put(cmd)
        rsp = self.etp.read_rsp()
        value = struct.unpack('<I', rsp)[0]
        return value



"""
Copyright (C) 2024, Jabez Winston C

Embedded Tester Protocol Library - I2C

Author  : Jabez Winston C <jabezwinston@gmail.com>
License : MIT
Date    : 19-Sep-2024

"""

import struct

class I2C:
    code = 5
    ops = {
        'info': 0,
        'init': 1,
        'scan': 2,
        'read': 3,
        'write': 4,
        'read_reg': 5,
        'write_reg': 6
    }

    info_cmds = {
        'bus_count': 0x01,
        'bus_speeds': 0x02,
        'pins': 0x03
    }

    def __init__(self, etp):
        self.etp = etp

    """
    Query I2C pin information, bus counts and supported speeds.

    """

    def get_info(self):
        i2c_info_cmd = self.code << 8 | self.ops['info']
        # Get the number of I2C buses
        cmd = self.etp.frame_packet(i2c_info_cmd, struct.pack('<B', self.info_cmds['bus_count']))
        self.etp.cmd_queue.put(cmd)
        rsp = self.etp.read_rsp()
        bus_count = struct.unpack('<B', rsp)[0]

        # Get the I2C bus speeds
        cmd = self.etp.frame_packet(i2c_info_cmd, struct.pack('<B', self.info_cmds['bus_speeds']))
        self.etp.cmd_queue.put(cmd)
        rsp = self.etp.read_rsp()
        speed_count = struct.unpack('<B', rsp[0:1])[0]
        bus_speeds = list(struct.unpack('<' + 'H'*speed_count, rsp[1:]))

        # Get the I2C pins
        cmd = self.etp.frame_packet(i2c_info_cmd, struct.pack('B', self.info_cmds['pins']))
        self.etp.cmd_queue.put(cmd)
        rsp = self.etp.read_rsp()

        # Use iter to unpack the I2C pins as a list of tuples
        pins = list(struct.iter_unpack('<BBB', rsp))

        pin_info = []

        for pin in pins:
            pin_info.append({'port': chr(pin[0]), 'pins': {'sda' : pin[1], 'scl' : pin[2]}})

        return {"bus_count": bus_count, "bus_speeds": bus_speeds, "info": pin_info}
    
    """
    Initialize I2C bus speed
    
    """
    
    def init(self, bus, speed):
        i2c_init_cmd = self.code << 8 | self.ops['init']
        cmd = self.etp.frame_packet(i2c_init_cmd, struct.pack('<BH', bus, speed))
        self.etp.cmd_queue.put(cmd)
        rsp = self.etp.read_rsp()
        return rsp
    
    """
    Scan I2C bus for devices

    """
    
    def scan(self, bus=0, start_addr=0, end_addr=0x7F):
        i2c_scan_cmd = self.code << 8 | self.ops['scan']
        cmd = self.etp.frame_packet(i2c_scan_cmd, struct.pack('<BHH', bus, start_addr, end_addr))
        self.etp.cmd_queue.put(cmd)
        rsp = self.etp.read_rsp()

        addr_count = rsp[0]
        addr_list = list(struct.unpack('<H'*addr_count, rsp[1:]))
        return addr_list
    
    """
    Read data from I2C device

    """
    
    def read(self, bus, addr, num_bytes):
        i2c_read_cmd = self.code << 8 | self.ops['read']
        cmd = self.etp.frame_packet(i2c_read_cmd, struct.pack('<BHB', bus, addr, num_bytes))
        self.etp.cmd_queue.put(cmd)
        rsp = self.etp.read_rsp()
        len = rsp[0]
        data = list(rsp[1:])
        return data
    
    """
    Write data to I2C device

    """
    
    def write(self, bus, addr, data):
        i2c_write_cmd = self.code << 8 | self.ops['write']
        length = len(data)
        cmd = self.etp.frame_packet(i2c_write_cmd, struct.pack('<BHB', bus, addr, length) + bytes(data))
        self.etp.cmd_queue.put(cmd)
        rsp = self.etp.read_rsp()
        return rsp[0]
    
    """
    Read data from I2C device register

    """
    
    def read_reg(self, bus, addr, reg, num_bytes):
        i2c_read_reg_cmd = self.code << 8 | self.ops['read_reg']
        cmd = self.etp.frame_packet(i2c_read_reg_cmd, struct.pack('<BHHB', bus, addr, reg, num_bytes))
        self.etp.cmd_queue.put(cmd)
        rsp = self.etp.read_rsp()
        len = rsp[0]
        data = list(rsp[1:])
        return data
    
    """
    Write data to I2C device register
    
    """
    
    def write_reg(self, bus, addr, reg, data):
        i2c_write_reg_cmd = self.code << 8 | self.ops['write_reg']
        length = len(data)
        cmd = self.etp.frame_packet(i2c_write_reg_cmd, struct.pack('<BHHB', bus, addr, reg, length) + bytes(data))
        self.etp.cmd_queue.put(cmd)
        rsp = self.etp.read_rsp()
        return rsp[0]

        


"""
Copyright (C) 2024, Jabez Winston C

Embedded Tester Protocol Library

Author  : Jabez Winston C <jabezwinston@gmail.com>
License : MIT
Date    : 13-Sep-2024

"""

import serial
import struct
import queue
import threading
import time

from .gpio import GPIO
from .adc import ADC
from .i2c import I2C
from .pwm import PWM

class ETP:
    general_ops = {
        'fw_info': 0,
        'reset': 1,
        'get_supported_ops': 2
    }

    fw_info_cmds = {
        'version': 1,
        'build_date': 2,
        'hw_type': 3
    }

    payload_types = {
        'cmd': 1,
        'data': 2,
        'rsp': 3,
        'event': 4
    }

    def __init__(self, **kwargs):
        if 'transport' in kwargs:
            self.transport = kwargs['transport']

        if 'port' in kwargs:
            self.port = kwargs['port']

        if 'baudrate' in kwargs:
            self.baudrate = kwargs['baudrate']
        else:
            self.baudrate = 115200

        self.ser = None
        self.ser_open = False
        self.rsp = []
        self.cmd_queue = queue.Queue()
        self.rsp_queue = queue.Queue()
        self.lock = threading.Lock()

    def reader_thread(self):
        while self.ser_open:
            try:
                if (self.ser.in_waiting > 0):
                    self.rsp.extend(self.ser.read(self.ser.in_waiting))
                    if len(self.rsp) > 0:
                        length = self.rsp[0]
                        self.rsp.extend(self.ser.read(length - len(self.rsp)))
                        self.rsp_queue.put(self.rsp[0: self.rsp[0]])
                        self.rsp = self.rsp[self.rsp[0]:]
                else:
                    time.sleep(0.001)
            except Exception as e:
                print(f"Error reading from serial port: {e}")
                break

    def writer_thread(self):
        while self.ser_open:
            try:
                if not self.cmd_queue.empty():
                    cmd = self.cmd_queue.get()
                    self.ser.write(cmd)
                else:
                    time.sleep(0.001)
            except Exception as e:
                print(f"Error writing to serial port: {e}")
                break


    def mask_to_bits(self, mask, bit_count):
        bits = []
        for i in range(bit_count):
            if mask & (1 << i):
                bits.append(i)
        return bits

    def frame_packet(self, cmd, data=b''):
        self.transaction_id = 0
        packet = struct.pack('<BBHH', len(data) + 6, self.payload_types['cmd'], self.transaction_id, cmd) + bytes(data)
        return packet

    def read_rsp(self):
        try:
            rsp = self.rsp_queue.get(timeout=1)
            rsp = bytearray(rsp[6:])
            return rsp
        except Exception as e:
            return None

    def open(self):
        if self.transport == 'serial':
            self.ser = serial.Serial(self.port, self.baudrate, timeout=1)
            self.ser.timeout = 0.1
            self.ser_open = True

        time.sleep(0.2)

        self.reader_thread_handle = threading.Thread(target=self.reader_thread)
        self.reader_thread_handle.start()
        self.writer_thread_handle = threading.Thread(target=self.writer_thread)
        self.writer_thread_handle.start()

        self.gpio = GPIO(self)
        self.adc = ADC(self)
        self.i2c = I2C(self)
        self.pwm = PWM(self)

    """
    Get Firmware Information
    
    """
    def get_fw_info(self):
        self.cmd_queue.put(self.frame_packet(self.general_ops['fw_info'], [self.fw_info_cmds['version']]))
        rsp = self.read_rsp()
        version = struct.unpack('<BBB', rsp[:3])
        version_str = '.'.join([str(x) for x in version])

        self.cmd_queue.put(self.frame_packet(self.general_ops['fw_info'], [self.fw_info_cmds['build_date']]))
        rsp = self.read_rsp()
        year, month, day = struct.unpack('<HBB', rsp[:4])
        hr, min, sec = struct.unpack('<BBB', rsp[4:])
        build_date_str = f"{day}-{month}-{year},{hr}:{min}:{sec}"

        self.cmd_queue.put(self.frame_packet(self.general_ops['fw_info'], [self.fw_info_cmds['hw_type']]))
        rsp = self.read_rsp()
        hw_type = rsp.decode('utf-8')
        return {'version': version_str, 'build_date': build_date_str, 'hw_type': hw_type}
    
    """
    Reset the device
        
    """
    def reset(self):
        self.cmd_queue.put(self.frame_packet(self.general_ops['reset'], [1]))
        rsp = self.read_rsp()
        return rsp
    
    """
    Get Supported Operations

    """
    def get_supported_ops(self, start_op = 0, end_op = 0xFFFF):
        supported_ops = []
        sub_cmd = [start_op & 0xFF, start_op >> 8, end_op & 0xFF, end_op >> 8]
        p = self.frame_packet(self.general_ops['get_supported_ops'], sub_cmd)
        self.cmd_queue.put(p)
        rsp = self.read_rsp()

        total_ops = struct.unpack('<H', rsp[:2])[0]
        report_ops = rsp[2]

        ops = rsp[3:]

        for i in range(0, len(ops), 2):
            supported_ops.append(ops[i] | ops[i + 1] << 8)

        next_op = supported_ops[-1] + 1

        while total_ops > report_ops:
            sub_cmd = [next_op & 0xFF, next_op >> 8, end_op & 0xFF, end_op >> 8]
            p = self.frame_packet(self.general_ops['get_supported_ops'], sub_cmd)
            self.cmd_queue.put(p)
            rsp = self.read_rsp()

            report_ops += rsp[2]
            ops = rsp[3:]

            for i in range(0, len(ops), 2):
                supported_ops.append(ops[i] | ops[i + 1] << 8)

        return supported_ops

    def close(self):
        self.ser_open = False
        self.reader_thread_handle.join()
        self.writer_thread_handle.join()
        self.ser.close()



"""
Copyright (C) 2024, Jabez Winston C

Embedded Tester Protocol Library - GPIO

Author  : Jabez Winston C <jabezwinston@gmail.com>
License : MIT
Date    : 17-Sep-2024

"""

import struct

class GPIO:
    code = 1
    ops = {
        'info': 0,
        'init': 1,
        'read': 2,
        'write': 3
    }

    info_cmds = {
        'port_count': 0x01,
        'pin_count' : 0x02,
        'ports'     : 0x03,
        'pins'      : 0x04
    }

    def __init__(self, etp):
        self.etp = etp

    """
    Get GPIO port and pin information
    
    """

    def get_info(self):
        gpio_info_cmd = self.code << 8 | self.ops['info']
        # Get the number of GPIO ports
        cmd = self.etp.frame_packet(gpio_info_cmd, struct.pack('B', self.info_cmds['port_count']))
        self.etp.cmd_queue.put(cmd)
        rsp = self.etp.read_rsp()
        port_count = struct.unpack('B', rsp)[0]

        # Get the GPIO pins
        cmd = self.etp.frame_packet(gpio_info_cmd, struct.pack('B', self.info_cmds['pins']))
        self.etp.cmd_queue.put(cmd)
        rsp = self.etp.read_rsp()

        # Use iter to unpack the GPIO pins as a list of tuples
        pins = list(struct.iter_unpack('<BI', rsp))

        pin_info = []

        for pin in pins:
            pin_info.append({'port': chr(pin[0]), 'pins': self.etp.mask_to_bits(pin[1], 32)})

        return {"port_count": port_count, "info": pin_info}

    def decode_gpio_pin(self, pin):
        port = ''
        pin_num = 0

        if pin[0] == 'P' and pin[1] >= 'A' and pin[1] <= 'Z' and len(pin) >= 3:
            port = pin[1]
            pin_num = int(pin[2:])
        elif pin[0] == 'P' and pin[1] >= '0' and pin[1] <= '9' and pin[2] == '.' and pin[3] >= '0' and pin[3] <= '9' and len(pin) >= 4:
            port = pin[1]
            pin_num = int(pin[3:])
        elif pin[0] == '_' and pin[1] >= '0' and pin[1] <= '9' and len(pin) >= 2:
            port = pin[0]
            pin_num = int(pin[1:])
        elif pin[1] >= '0' and pin[1] <= '9' and len(pin) >= 2:
            port = pin[0]
            pin_num = int(pin[1:])

        return port, pin_num
    
    def encode_gpio_pin(self, port, pin_num):
        if port >= ord('A') and port <= ord('Z'):
            return f"P{chr(port)}{pin_num}"
        elif port >= ord('a') and port <= ord('z'):
            return f"{chr(port)}{pin_num}"
        elif port == ord('_'):
            return f"_{pin_num}"
        
    """
    Initialize GPIO pins

    """

    def init(self, pin_list):
        gpio_init_cmd = self.code << 8 | self.ops['init']
        dir_mask = 0
        dir_val = 0
        pull_mask = 0
        pull_val = 0
        int_mask = 0
        int_val = 0
        for pin, config in pin_list.items():
            port, pin_num = self.decode_gpio_pin(pin)
            # GPIO dir.
            # Check for "mode" key in config
            if 'mode' in config:
                dir_mask |= (1 << pin_num)
                if config['mode'] == 'input':
                    dir_val |= (1 << pin_num)

            if 'type' in config:
                pull_mask |= (1 << pin_num)
                if config['type'] == 'pull_up':
                    pull_val |= (1 << pin_num)

            if 'interrupt' in config:
                int_mask |= (1 << pin_num)
                if config['interrupt'] == 'rising_edge':
                    int_val |= (1 << pin_num * 2)
                elif config['interrupt'] == 'falling_edge':
                    int_val |= (2 << pin_num * 2)
                elif config['interrupt'] == 'both_edges':
                    int_val |= (3 << pin_num * 2)

        cmd = self.etp.frame_packet(gpio_init_cmd, struct.pack('<BIIIIIQ', ord(port), dir_mask, dir_val, pull_mask, pull_val, int_mask, int_val))
        self.etp.cmd_queue.put(cmd)
        self.etp.read_rsp()

    """
    Read GPIO pins provided as list
    """

    def read(self, pin_list):
        pin_read = dict()
        gpio_read_cmd = self.code << 8 | self.ops['read']
        pin_mask = 0
        for pin in pin_list:
            port, pin_num = self.decode_gpio_pin(pin)
            # Port mask, pin mask (32 bits)
            pin_mask |= (1 << pin_num)

        cmd = self.etp.frame_packet(gpio_read_cmd, struct.pack('<BI', ord(port), pin_mask))
        self.etp.cmd_queue.put(cmd)
        rsp = self.etp.read_rsp()
        port, pin_mask, port_value = struct.unpack('<BII', rsp)
        pins = self.etp.mask_to_bits(pin_mask, 32)
        for pin in pins:
            pin_read[self.encode_gpio_pin(port, pin)] = (port_value & (1 << pin)) >> pin

        return pin_read
    
    """
    Write GPIO pins provided as dictionary
    
    """

    def write(self, pin_state):
        gpio_write_cmd = self.code << 8 | self.ops['write']
        port = ''
        port_mask = 0
        port_mask_val = 0
        for pin, state in pin_state.items():
            port, pin_num = self.decode_gpio_pin(pin)
            port_mask |= (1 << pin_num)
            if state:
                port_mask_val |= (1 << pin_num)
        
        cmd = self.etp.frame_packet(gpio_write_cmd, struct.pack('<BII', ord(port), port_mask, port_mask_val))
        self.etp.cmd_queue.put(cmd)



"""
Copyright (C) 2024, Jabez Winston C

Embedded Tester Protocol Library - ADC

Author  : Jabez Winston C <jabezwinston@gmail.com>
License : MIT
Date    : 13-Sep-2024

"""

from .version import VERSION
from .lib import *

__version__ = VERSION
__author__  = "Jabez Winston C"
__license__ = "MIT"
__all__     = ['ETP', 'GPIO', 'ADC', 'I2C', 'PWM']
__package__ = "etplib"



"""
Copyright (C) 2024, Jabez Winston C

Embedded Tester Protocol Library - PWM

Author  : Jabez Winston C <jabezwinston@gmail.com>
License : MIT
Date    : 19-Sep-2024

"""

import struct

class PWM:
    code = 3
    ops = {
        'info': 0,
        'init': 1,
        'ctrl': 2
    }

    def __init__(self, etp):
        self.etp = etp

    """
    Query PWM information
    
    """

    def get_info(self):
        pwm_info_cmd = self.code << 8 | self.ops['info']
        cmd = self.etp.frame_packet(pwm_info_cmd)
        self.etp.cmd_queue.put(cmd)
        rsp = self.etp.read_rsp()
        num_pwm, max_freq, freq_unit, port_count = struct.unpack('<BHBB', rsp[0:5])
        pwm_ports = list(struct.iter_unpack('<BI', rsp[5:]))
        pwm_info = []
        for port in pwm_ports:
            pwm_info.append({'port': chr(port[0]), 'pins': self.etp.mask_to_bits(port[1], 32)})

        if freq_unit == 1:
            max_freq *= 1
        elif freq_unit == 2:
            max_freq *= 1000

        return {
            'num_pwm': num_pwm,
            'max_freq': max_freq,
            'port_count': port_count,
            'ports': pwm_info
        }
    
    """
    Enable/Disable PWM pins

    """

    def init(self, pin_list):
        pwm_init_cmd = self.code << 8 | self.ops['init']
        pwm_pin_mask = 0
        pwm_enable_mask = 0
        for pin in pin_list.keys():
            port, pin_num = self.etp.gpio.decode_gpio_pin(pin)
            if (pin_list[pin] == True):
                pwm_enable_mask |= (1 << pin_num)
            pwm_pin_mask |= (1 << pin_num)

        cmd = self.etp.frame_packet(pwm_init_cmd, struct.pack('<BII', ord(port), pwm_pin_mask, pwm_enable_mask))
        self.etp.cmd_queue.put(cmd)

    """
    Control PWM pins
    
    """

    def ctrl(self, pin_str, duty_cycle, freq = 1000):
        pwm_ctrl_cmd = self.code << 8 | self.ops['ctrl']
        port, pin_num = self.etp.gpio.decode_gpio_pin(pin_str)
        freq_unit = 0
        if freq < 65535:
            freq_unit = 0
        else:
            freq_unit = 1

        duty_cycle = int(duty_cycle * 100)
        cmd = self.etp.frame_packet(pwm_ctrl_cmd, struct.pack('<BBHBH', ord(port), pin_num, freq, freq_unit, duty_cycle))
        self.etp.cmd_queue.put(cmd)
